
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "t_m.h"
#include "xyz.h"
#include "sgif.h"

extern void	Error(),ColorFile();

#define WORD	unsigned short
#define BYTE	unsigned char

typedef struct {
		WORD	hi,vi,hj,vj,hk,vk,d;
		BYTE	R,G,B;
#ifdef	__BORLANDC__
/*
		BYTE	pad[3];
*/
#endif
	} Trigon;

typedef struct {
		size_t	trigons;
		Trigon	*trigon;
	} Surface;

static int	D;
static Surface	*S;
static char	*tmae="trigon memory allocation error.";

static void	StoreTrigon(hi,vi,hj,vj,hk,vk,c)
int		hi,vi,hj,vj,hk,vk;
long		c;
{
	size_t	o=S->trigons++;
	Trigon	*t;

	if ((S->trigon=(Trigon *)
	     realloc(S->trigon,S->trigons*sizeof(Trigon)))==NULL) Error(tmae);

	t=S->trigon+o; t->hi=(WORD)hi; t->vi=(WORD)vi;
		       t->hj=(WORD)hj; t->vj=(WORD)vj;
		       t->hk=(WORD)hk; t->vk=(WORD)vk; t->d=(WORD)D;
						       t->R=(BYTE)(c    &0xff);
						       t->G=(BYTE)(c>>8 &0xff);
						       t->B=(BYTE)(c>>16&0xff);
}

#define PATH_LEN	1024

static FILE	*CreateVM(prefix,suffix)
char		*prefix,*suffix;
{
	char	path[PATH_LEN];
	FILE	*file;

	(void)sprintf(path,"%s.%s_m",prefix,suffix);
	if ((file=fopen(path,"w+b"))==NULL) {
	    (void)fputs(path,stderr); Error(" : VM file not open.");
	}
	return file;
}

static FILE	*FVM;

static void	DumpVM(hi,vi,hj,vj,hk,vk,c)
int		hi,vi,hj,vj,hk,vk;
long		c;
{
	Trigon	t;

	t.hi=(WORD)hi; t.vi=(WORD)vi;
	t.hj=(WORD)hj; t.vj=(WORD)vj;
	t.hk=(WORD)hk; t.vk=(WORD)vk; t.d=(WORD)D;
				      t.R=(BYTE)(c    &0xff);
				      t.G=(BYTE)(c>>8 &0xff);
				      t.B=(BYTE)(c>>16&0xff);
	if (fwrite(&t,sizeof(Trigon),(size_t)1,FVM)!=1)
	    Error("trigon data write error.");

	++S->trigons;
}

#define ALLOC(type,noe)	(type *)malloc((size_t)(noe)*sizeof(type))

static void	RestoreVM(file,s)
FILE		*file;
Surface		*s;
{
	if ((s->trigon=ALLOC(Trigon,s->trigons))==NULL) Error(tmae);

	if (fseek(file,0L,SEEK_SET) ||
	    fread(s->trigon,sizeof(Trigon),s->trigons,file)!=s->trigons)
	    Error("trigon data read error.");

	(void)fclose(file);
}

static void	SI_Surface(path,Nx,Ny,Nz,solid,R,G,B,VM,mx,px,my,py,mz,pz,gray)
char		**path,*solid,*VM;
int		Nx,Ny,Nz,*gray;
BYTE		*R,*G,*B;
Surface		*mx,*px,*my,*py,*mz,*pz;
{
	Cell	***image,**cell,**slice;
	char	**back,
		*imae="image memory allocation error.";
	int	x1,Y1,z1,x2,y2,z2,z,nx,ny,y,x,pv,N,v,h,h1,v1,h2,v2;
	long	**fore;
	void	(*store)();
	FILE	*Fmx,*Fpx,*Fmy,*Fpy,*Fmz,*Fpz;

	if ((image=ALLOC(Cell **,Nz))==NULL) Error(imae);

	x1=Nx; Y1=Ny; z1=Nz; x2=y2=z2=(-1);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL);

	    if (nx>=Nx && ny>=Ny)
		image[z]=slice=cell;
	    else {
		if ((image[z]=slice=ALLOC(Cell *,Ny))==NULL) Error(imae);

		for (y=0; y<Ny; y++) {
		    if ((slice[y]=ALLOC(Cell,Nx))==NULL) Error(imae);

		    for (x=0; x<Nx; x++)
			slice[y][x]=(x<nx && y<ny)?cell[y][x]:0;
		}
		for (y=ny-1; y>=0; y--) free(cell[y]); free(cell);
	    }
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++)
		if (solid[pv=slice[y][x]]) {
		    if (x<x1) x1=x; if (x>x2) x2=x;
		    if (y<Y1) Y1=y; if (y>y2) y2=y;
		    if (z<z1) z1=z; if (z>z2) z2=z;

		    if (*gray && (R[pv]!=G[pv] || G[pv]!=B[pv])) *gray=0;
		}
	}
	if (x1>x2 || Y1>y2 || z1>z2) Error("no solid object.");

	N=(Nx>Ny)?(Nx>Nz)?Nx:Nz:(Ny>Nz)?Ny:Nz;

	InitTriangulation(N,N,&back,&fore);

	store=(VM==NULL)?StoreTrigon:DumpVM;

#define SETUP(Nh,Nv,surface,file,extension)	\
	for (v=0; v<Nv; v++) for (h=0; h<Nh; h++) back[v][h]=1; \
	S=surface; S->trigons=0; \
	if (VM==NULL) S->trigon=NULL; else FVM=file=CreateVM(VM,extension)

#define TRIANGULATION(H,H1,H2,V,V1,V2,d)	{ \
	h1=v1=N; h2=v2=(-1); \
	for (V=V1; V<=V2; V++) \
	for (H=H1; H<=H2; H++) \
	    if (!(back[V][H]=(char)(((back[V][H]&1)?0:2)| \
				    ((solid[pv=image[z][y][x]])?0:1)))) { \
		fore[V][H]=(long)R[pv]|(long)G[pv]<<8|(long)B[pv]<<16; \
		if (H<h1) h1=H; if (H>h2) h2=H; \
		if (V<v1) v1=V; if (V>v2) v2=V; \
	    } \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,fore,store); \
	} \
}
	SETUP(Ny,Nz,mx,Fmx,"mx");
	for (x=x1; x<=x2; x++) TRIANGULATION(y,Y1,y2,z,z1,z2,x  )
	SETUP(Ny,Nz,px,Fpx,"px");
	for (x=x2; x>=x1; x--) TRIANGULATION(y,Y1,y2,z,z1,z2,x+1)
	SETUP(Nz,Nx,my,Fmy,"my");
	for (y=Y1; y<=y2; y++) TRIANGULATION(z,z1,z2,x,x1,x2,y  )
	SETUP(Nz,Nx,py,Fpy,"py");
	for (y=y2; y>=Y1; y--) TRIANGULATION(z,z1,z2,x,x1,x2,y+1) 
	SETUP(Nx,Ny,mz,Fmz,"mz");
	for (z=z1; z<=z2; z++) TRIANGULATION(x,x1,x2,y,Y1,y2,z  ) 
	SETUP(Nx,Ny,pz,Fpz,"pz");
	for (z=z2; z>=z1; z--) TRIANGULATION(x,x1,x2,y,Y1,y2,z+1)

	TermTriangulation(N,back,fore);

	for (z=Nz-1; z>=0; z--) {
	    for (y=Ny-1; y>=0; y--) free(image[z][y]); free(image[z]);
	}
	free(image);

	if (VM!=NULL) {
	    RestoreVM(Fmx,mx);
	    RestoreVM(Fpx,px);
	    RestoreVM(Fmy,my);
	    RestoreVM(Fpy,py);
	    RestoreVM(Fmz,mz);
	    RestoreVM(Fpz,pz);
	}
}

static BYTE	**r,**g,**b,pvR,pvG,pvB;
static int	H,V;

static void	PlotPV(h,v)
int		h,v;
{
	h+=H; v=V-v; r[v][h]=pvR; g[v][h]=pvG; b[v][h]=pvB;
}

static double	Gamma,bias;

#define SHADE(shade)	(bias+(255.-bias)*pow(shade,1./Gamma))/255.

#ifndef	NSS
static void	PlotShade(h,v,shade)
int		h,v;
double		shade;
{
	double	I=SHADE(shade);

	h+=H; v=V-v; r[v][h]=(BYTE)(I*(double)r[v][h]+0.5);
		     g[v][h]=(BYTE)(I*(double)g[v][h]+0.5);
		     b[v][h]=(BYTE)(I*(double)b[v][h]+0.5);
}
#endif

#define TEXT_LEN	256

static void	Surface_BEV(Nx,Ny,Nz,mx,px,my,py,mz,pz,
			    size,uX,uY,uZ,bgR,bgG,bgB,fgR,fgG,fgB,path,gray)
int		Nx,Ny,Nz,size,bgR,bgG,bgB,fgR,fgG,fgB;
Surface		*mx,*px,*my,*py,*mz,*pz;
double		uX,uY,uZ;
char		*path;
BYTE		*gray;
{
	double	Wx,Wy,Wz;
	int	v;
	FILE	*file;
	char	text[TEXT_LEN],
		*fmae="frame memory allocation error.";

	if (size) {
	    Wx=(double)Nx;
	    Wy=(double)Ny;
	    Wz=(double)Nz; uX=
			   uY=
			   uZ=(double)size/sqrt(Wx*Wx+Wy*Wy+Wz*Wz); Wx*=uX;
								    Wy*=uY;
								    Wz*=uZ;
	}
	else {
	    Wx=uX*(double)Nx;
	    Wy=uY*(double)Ny;
	    Wz=uZ*(double)Nz; size=(int)ceil(sqrt(Wx*Wx+Wy*Wy+Wz*Wz));
	}
	if (gray==NULL) {
	    if ((r=ALLOC(BYTE *,size))==NULL ||
		(g=ALLOC(BYTE *,size))==NULL ||
		(b=ALLOC(BYTE *,size))==NULL) Error(fmae);

	    for (v=0; v<size; v++)
		if ((r[v]=ALLOC(BYTE,size))==NULL ||
		    (g[v]=ALLOC(BYTE,size))==NULL ||
		    (b[v]=ALLOC(BYTE,size))==NULL) Error(fmae);
	}
	else {
	    if ((r=g=b=ALLOC(BYTE *,size))==NULL) Error(fmae);

	    for (v=0; v<size; v++)
		if ((r[v]=ALLOC(BYTE,size))==NULL) Error(fmae);
	}
	if (*path=='-')
	    file=stdout;
	else
	    if ((file=fopen(path,"wb"))==NULL) Error("GIF file not open.");

	InitStoreGIF(file,size,size,gray,gray,gray);

	while (fgets(text,TEXT_LEN,stdin)!=NULL)
{
	double	bLo,bLa,sLo,sLa,lo,la,Wh,Wv,side,x,y,z;
	int	cnt,idx,h,n;
	Trigon	*t;

#ifdef	NSS
	double	I;
#endif

	switch(sscanf(text,"%lf %lf %lf %lf %d",&bLo,&bLa,&sLo,&sLa,&cnt)) {
	    case 2: sLo=sLa=0.; cnt=1; break;
	    case 5: break;
	    default: continue;
	}
	for (idx=0; idx<cnt; idx++) {
	    lo=(bLo+sLo*(double)idx)*M_PI/180.;
	    la=(bLa+sLa*(double)idx)*M_PI/180.;

	    XYZWidth(lo,la,0.,0.,0.,Wx,Wy,Wz,&Wh,&Wv);
	    h=(int)ceil(Wh); H=(size-h)/2;
	    v=(int)ceil(Wv); V=(size+v)/2-1;
	    if (!XYZInit(h,v,lo,la,0.,0.,0.,Wx,Wy,Wz))
		Error("rendering memory allocation error.");

	    for (v=0; v<size; v++)
	    for (h=0; h<size; h++) {
		r[v][h]=(BYTE)bgR; g[v][h]=(BYTE)bgG; b[v][h]=(BYTE)bgB;
	    }

#ifndef	NSS
#define COLOR_I(shade)
#define COLOR_RGB()	pvR=t->R; pvG=t->G; pvB=t->B
#else
#define COLOR_I(shade)	I=SHADE(shade);
#define COLOR_RGB()	pvR=(BYTE)(I*(double)t->R+0.5); \
			pvG=(BYTE)(I*(double)t->G+0.5); \
			pvB=(BYTE)(I*(double)t->B+0.5)
#endif

	    if ((side=XYZSide(lo,la,0.,0.,0.,0.,1.,0.,0.,0.,1.))!=0.) {
		if (side<0.) {
		    n=mx->trigons; t=mx->trigon; COLOR_I(-side)
		}
		else {
		    n=px->trigons; t=px->trigon; COLOR_I( side)
		}
		for (; n; n--, t++) {
		    x=uX*(double)t->d; COLOR_RGB();
		    XYZFill(x,uY*(double)t->hi,uZ*(double)t->vi,
			    x,uY*(double)t->hj,uZ*(double)t->vj,
			    x,uY*(double)t->hk,uZ*(double)t->vk,PlotPV);
		}
	    }
	    if ((side=XYZSide(lo,la,0.,0.,0.,0.,0.,1.,1.,0.,0.))!=0.) {
		if (side<0.) {
		    n=my->trigons; t=my->trigon; COLOR_I(-side)
		}
		else {
		    n=py->trigons; t=py->trigon; COLOR_I( side)
		}
		for (; n; n--, t++) {
		    y=uY*(double)t->d; COLOR_RGB();
		    XYZFill(uX*(double)t->vi,y,uZ*(double)t->hi,
			    uX*(double)t->vj,y,uZ*(double)t->hj,
			    uX*(double)t->vk,y,uZ*(double)t->hk,PlotPV);
		}
	    }
	    if ((side=XYZSide(lo,la,0.,0.,0.,1.,0.,0.,0.,1.,0.))!=0.) {
		if (side<0.) {
		    n=mz->trigons; t=mz->trigon; COLOR_I(-side)
		}
		else {
		    n=pz->trigons; t=pz->trigon; COLOR_I( side)
		}
		for (; n; n--, t++) {
		    z=uZ*(double)t->d; COLOR_RGB();
		    XYZFill(uX*(double)t->hi,uY*(double)t->vi,z,
			    uX*(double)t->hj,uY*(double)t->vj,z,
			    uX*(double)t->hk,uY*(double)t->vk,z,PlotPV);
		}
	    }
#ifndef	NSS
	    XYZShade(lo,la,PlotShade);
#endif
	    if (fgR>=0 && fgG>=0 && fgB>=0) {
		pvR=(BYTE)fgR; pvG=(BYTE)fgG; pvB=(BYTE)fgB;
		XYZDraw(0.,0.,0.,Wx,0.,0.,PlotPV);
		XYZDraw(Wx,0.,0.,Wx,Wy,0.,PlotPV);
		XYZDraw(Wx,Wy,0.,0.,Wy,0.,PlotPV);
		XYZDraw(0.,Wy,0.,0.,0.,0.,PlotPV);
		XYZDraw(0.,0.,Wz,Wx,0.,Wz,PlotPV);
		XYZDraw(Wx,0.,Wz,Wx,Wy,Wz,PlotPV);
		XYZDraw(Wx,Wy,Wz,0.,Wy,Wz,PlotPV);
		XYZDraw(0.,Wy,Wz,0.,0.,Wz,PlotPV);
		XYZDraw(0.,0.,0.,0.,0.,Wz,PlotPV);
		XYZDraw(Wx,0.,0.,Wx,0.,Wz,PlotPV);
		XYZDraw(Wx,Wy,0.,Wx,Wy,Wz,PlotPV);
		XYZDraw(0.,Wy,0.,0.,Wy,Wz,PlotPV);
	    }
	    XYZTerm();

	    if (gray==NULL)
		StoreGIF_RGB(file,0,0,size,size,r,g,b);
	    else
		StoreGIF(file,0,0,size,size,NULL,NULL,NULL,r);
	}
}
	(void)TermStoreGIF(file); if (file!=stdout) (void)fclose(file);

	if (gray==NULL) {
	    for (v=size-1; v>=0; v--) {
		free(b[v]); free(g[v]); free(r[v]);
	    }
	    free(b); free(g); free(r);
	}
	else {
	    for (v=size-1; v>=0; v--) free(r[v]); free(r);
	}
}

#define PV	(int)((Cell)(~0))

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**path,solid[PV+1];
	int	Nx,Ny,Nz,size,bgR,bgG,bgB,fgR,fgG,fgB,gray;
	BYTE	R[PV+1],G[PV+1],B[PV+1];
	double	uX,uY,uZ;
	Surface	mx,px,my,py,mz,pz;

	if (argc<14 || argc>17)
	    Error("usage : si_m_bev directory nameFile colorFile {VM}\n"
		  "                 size|(unitX unitY unitZ) gamma bias "
				   "bgR bgG bgB fgR fgG fgB GIF");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	ColorFile(argv[3],solid,R,G,B);

	if (argc==16 || argc==17) {
	    if ((uX=atof(argv[argc-12]))<=0. ||
		(uY=atof(argv[argc-11]))<=0. ||
		(uZ=atof(argv[argc-10]))<=0.) Error("bad unit.");

	    size=0;
	}
	else
	    if ((size=atoi(argv[argc-10]))<=0) Error("bad size.");

	if ((Gamma=atof(argv[argc-9]))<=0.) Error("bad gamma.");
	if ((bias=atof(argv[argc-8]))<0. || bias>255.) Error("bad bias.");

	if ((bgR=atoi(argv[argc-7]))<0 || bgR>255 ||
	    (bgG=atoi(argv[argc-6]))<0 || bgG>255 ||
	    (bgB=atoi(argv[argc-5]))<0 || bgB>255) Error("bad background.");
	if ((fgR=atoi(argv[argc-4]))>255 ||
	    (fgG=atoi(argv[argc-3]))>255 ||
	    (fgB=atoi(argv[argc-2]))>255) Error("bad foreground.");

	if (fgR<0 || fgG<0 || fgB<0) fgR=fgG=fgB=(-1);

	gray=bgR==bgG && bgG==bgB && fgR==fgG && fgG==fgB;

	SI_Surface(path,Nx,Ny,Nz,solid,R,G,B,
		   (argc==15 || argc==17)?argv[4]:NULL,
		   &mx,&px,&my,&py,&mz,&pz,&gray);

	Surface_BEV(Nx,Ny,Nz,&mx,&px,&my,&py,&mz,&pz,
		    size,uX,uY,uZ,bgR,bgG,bgB,fgR,fgG,fgB,argv[argc-1],
		    (gray)?StoreGIF_Gray:NULL);

	return 0;
}
