
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "t_m.h"

extern void	Error(),ColorFile();

#define BYTE	unsigned char

static size_t	fwrite_reverse(data,size,memb,file)
BYTE		*data;
size_t		size,memb;
FILE		*file;
{
	int	m,s;

	for (m=0; m<memb; m++) {
	    data+=size;
	    for (s=1; s<=size; s++)
		if (putc((int)data[-s],file)==EOF) goto Break;
	}
Break:;
	return (size_t)m;
}

#define WORD	unsigned short

static double		uX=1.,uY=1.,uZ=1.;
static FILE		*STL;
static size_t		(*fWrite)();
static unsigned int	facets=0;

static void	Store(nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk,c)
int		nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk;
long		c;
{
	float	xyz[12];
	WORD	flag=(WORD)c;

	xyz[0]=   (float)nx; xyz[ 1]=   (float)ny; xyz[ 2]=   (float)nz;
	xyz[3]=uX*(float)xi; xyz[ 4]=uY*(float)yi; xyz[ 5]=uZ*(float)zi;
	xyz[6]=uX*(float)xj; xyz[ 7]=uY*(float)yj; xyz[ 8]=uZ*(float)zj;
	xyz[9]=uX*(float)xk; xyz[10]=uY*(float)yk; xyz[11]=uZ*(float)zk;
	if ((*fWrite)(xyz,sizeof(float),(size_t)12,STL)!=12 ||
	    (*fWrite)(&flag,sizeof(WORD),(size_t)1,STL)!=1)
	    Error("STL data write error.");

	++facets;
}

static int	D;

#define STORE(function,nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk)	\
	static void	function(hi,vi,hj,vj,hk,vk,c)	\
	int		hi,vi,hj,vj,hk,vk; \
	long		c; \
	{ \
		Store(nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk,c); \
	}

STORE(Mx,-1, 0, 0,D ,hi,vi,D ,hj,vj,D ,hk,vk)
STORE(Px, 1, 0, 0,D ,hi,vi,D ,hj,vj,D ,hk,vk)
STORE(My, 0,-1, 0,vi,D ,hi,vj,D ,hj,vk,D ,hk)
STORE(Py, 0, 1, 0,vi,D ,hi,vj,D ,hj,vk,D ,hk)
STORE(Mz, 0, 0,-1,hi,vi,D ,hj,vj,D ,hk,vk,D )
STORE(Pz, 0, 0, 1,hi,vi,D ,hj,vj,D ,hk,vk,D )

#define PV		(int)((Cell)(~0))
#define ALLOC(type,noe)	(type *)malloc((size_t)(noe)*sizeof(type))

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**path,solid[PV+1],**back,
		*mae="memory allocation error.";
	int	Nx,Ny,Nz,pv,x1,y1,z1,x2,y2,z2,z,nx,ny,y,x,N,v,h,h1,v1,h2,v2;
	WORD	flag[PV+1];
	Cell	***image,**cell,**slice;
	long	**fore;

	if (argc!=5 && argc!=8)
Error("usage : si_m_stl directory nameFile colorFile {unitX unitY unitZ} STL");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);
{
	BYTE	R[PV+1],G[PV+1],B[PV+1];

	ColorFile(argv[3],solid,R,G,B);

	for (pv=0; pv<=PV; pv++)
	    if (solid[pv])
		flag[pv]=(WORD)(0x8000|(((int)R[pv]&248)<<7)|
				       (((int)G[pv]&248)<<2)|
				       (((int)B[pv]&248)>>3));
}
	if ((image=ALLOC(Cell **,Nz))==NULL) Error(mae);

	x1=Nx; y1=Ny; z1=Nz; x2=y2=z2=(-1);
	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL);

	    if (nx>=Nx && ny>=Ny)
		image[z]=slice=cell;
	    else {
		if ((image[z]=slice=ALLOC(Cell *,Ny))==NULL) Error(mae);

		for (y=0; y<Ny; y++) {
		    if ((slice[y]=ALLOC(Cell,Nx))==NULL) Error(mae);

		    for (x=0; x<Nx; x++)
			slice[y][x]=(x<nx && y<ny)?cell[y][x]:0;
		}
		for (y=ny-1; y>=0; y--) free(cell[y]); free(cell);
	    }
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++)
		if (solid[slice[y][x]]) {
		    if (x<x1) x1=x; if (x>x2) x2=x;
		    if (y<y1) y1=y; if (y>y2) y2=y;
		    if (z<z1) z1=z; if (z>z2) z2=z;
		}
	}
	if (x1>x2 || y1>y2 || z1>z2) Error("no solid object.");

	N=(Nx>Ny)?(Nx>Nz)?Nx:Nz:(Ny>Nz)?Ny:Nz;

	InitTriangulation(N,N,&back,&fore);

	if (argc==8) {
	    uX=atof(argv[4]); uY=atof(argv[5]); uZ=atof(argv[6]);
	}
	if ((STL=fopen(argv[argc-1],"w+b"))==NULL) Error("STL file not open.");
{
	int	i;
	short	s=0x1234;
	char	*c=(char *)&s;

	for (i=0; i<80+sizeof(facets); i++)
	    if (putc(0,STL)==EOF) Error("STL header write error.");

	fWrite=(*c==(s&0xff))?fwrite:fwrite_reverse;
}

#define CLEAR(Nh,Nv)	for (v=0; v<Nv; v++) for (h=0; h<Nh; h++) back[v][h]=1

#define TRIANGULATION(H,H1,H2,V,V1,V2,d,store)	{ \
	h1=v1=N; h2=v2=(-1); \
	for (V=V1; V<=V2; V++) \
	for (H=H1; H<=H2; H++) \
	    if (!(back[V][H]=(char)(((back[V][H]&1)?0:2)| \
				    ((solid[pv=image[z][y][x]])?0:1)))) { \
		fore[V][H]=flag[pv]; if (H<h1) h1=H; if (H>h2) h2=H; \
				     if (V<v1) v1=V; if (V>v2) v2=V; \
	    } \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,fore,store); \
	} \
}
CLEAR(Ny,Nz); for (x=x1; x<=x2; x++) TRIANGULATION(y,y1,y2,z,z1,z2,x  ,Mx)
CLEAR(Ny,Nz); for (x=x2; x>=x1; x--) TRIANGULATION(y,y1,y2,z,z1,z2,x+1,Px)
CLEAR(Nz,Nx); for (y=y1; y<=y2; y++) TRIANGULATION(z,z1,z2,x,x1,x2,y  ,My)
CLEAR(Nz,Nx); for (y=y2; y>=y1; y--) TRIANGULATION(z,z1,z2,x,x1,x2,y+1,Py)
CLEAR(Nx,Ny); for (z=z1; z<=z2; z++) TRIANGULATION(x,x1,x2,y,y1,y2,z  ,Mz)
CLEAR(Nx,Ny); for (z=z2; z>=z1; z--) TRIANGULATION(x,x1,x2,y,y1,y2,z+1,Pz)

	if (fseek(STL,80L,SEEK_SET) ||
	    (*fWrite)(&facets,sizeof(facets),(size_t)1,STL)!=1)
	    Error("STL trailer write error.");

	(void)fclose(STL); TermTriangulation(N,back,fore); return 0;
}
