
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "t_s.h"
#include "xyz.h"
#include "sgif.h"

extern void	Error(),RangeList();

#define WORD	unsigned short

typedef struct {
		WORD	hi,vi,hj,vj,hk,vk,d;
#ifdef	__BORLANDC__
		WORD	pad;
#endif
	} Trigon;

typedef struct {
		size_t	trigons;
		Trigon	*trigon;
	} Surface;

static int	D;
static Surface	*S;
static char	*tmae="trigon memory allocation error.";

static void	StoreTrigon(hi,vi,hj,vj,hk,vk)
int		hi,vi,hj,vj,hk,vk;
{
	size_t	o=S->trigons++;
	Trigon	*t;

	if ((S->trigon=(Trigon *)
	     realloc(S->trigon,S->trigons*sizeof(Trigon)))==NULL) Error(tmae);

	t=S->trigon+o; t->hi=(WORD)hi; t->vi=(WORD)vi;
		       t->hj=(WORD)hj; t->vj=(WORD)vj;
		       t->hk=(WORD)hk; t->vk=(WORD)vk; t->d=(WORD)D;
}

#define PATH_LEN	1024

static FILE	*CreateVM(prefix,suffix)
char		*prefix,*suffix;
{
	char	path[PATH_LEN];
	FILE	*file;

	(void)sprintf(path,"%s.%s_s",prefix,suffix);
	if ((file=fopen(path,"w+b"))==NULL) {
	    (void)fputs(path,stderr); Error(" : VM file not open.");
	}
	return file;
}

static FILE	*FVM;

static void	DumpVM(hi,vi,hj,vj,hk,vk)
int		hi,vi,hj,vj,hk,vk;
{
	Trigon	t;

	t.hi=(WORD)hi; t.vi=(WORD)vi;
	t.hj=(WORD)hj; t.vj=(WORD)vj;
	t.hk=(WORD)hk; t.vk=(WORD)vk; t.d=(WORD)D;
	if (fwrite(&t,sizeof(Trigon),(size_t)1,FVM)!=1)
	    Error("trigon data write error.");

	++S->trigons;
}

#define ALLOC(type,noe)	(type *)malloc((size_t)(noe)*sizeof(type))

static void	RestoreVM(file,s)
FILE		*file;
Surface		*s;
{
	if ((s->trigon=ALLOC(Trigon,s->trigons))==NULL) Error(tmae);

	if (fseek(file,0L,SEEK_SET) ||
	    fread(s->trigon,sizeof(Trigon),s->trigons,file)!=s->trigons)
	    Error("trigon data read error.");

	(void)fclose(file);
}

static void	SI_Surface(path,Nx,Ny,Nz,solid,VM,mx,px,my,py,mz,pz)
char		**path,*solid,*VM;
int		Nx,Ny,Nz;
Surface		*mx,*px,*my,*py,*mz,*pz;
{
	char	***image,**i,m,**back,
		*imae="image memory allocation error.";
	int	x1,Y1,z1,x2,y2,z2,z,y,x,nx,ny,N,v,h,h1,v1,h2,v2;
	Cell	**cell;
	void	(*store)();
	FILE	*Fmx,*Fpx,*Fmy,*Fpy,*Fmz,*Fpz;

	if ((image=ALLOC(char **,(Nz+7)/8))==NULL) Error(imae);

	x1=Nx; Y1=Ny; z1=Nz; x2=y2=z2=(-1);

	for (z=0; z<Nz; z++) {
	    if (z%8==0) {
		if ((image[z/8]=i=ALLOC(char *,Ny))==NULL) Error(imae);

		for (y=0; y<Ny; y++) {
		    if ((i[y]=ALLOC(char,Nx))==NULL) Error(imae);

		    for (x=0; x<Nx; x++) i[y][x]=0;
		}
		m=1;
	    }
	    else
		m<<=1;

	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL);

	    for (y=0; y<ny && y<Ny; y++)
	    for (x=0; x<nx && x<Nx; x++)
		if (solid[cell[y][x]]) {
		    i[y][x]|=m; if (x<x1) x1=x; if (x>x2) x2=x;
				if (y<Y1) Y1=y; if (y>y2) y2=y;
				if (z<z1) z1=z; if (z>z2) z2=z;
		}

	    for (y=ny-1; y>=0; y--) free(cell[y]); free(cell);
	}
	if (x1>x2 || Y1>y2 || z1>z2) Error("no solid object.");

	N=(Nx>Ny)?(Nx>Nz)?Nx:Nz:(Ny>Nz)?Ny:Nz;

	InitTriangulation(N,N,&back);

	store=(VM==NULL)?StoreTrigon:DumpVM;

#define SETUP(Nh,Nv,surface,file,extension)	\
	for (v=0; v<Nv; v++) for (h=0; h<Nh; h++) back[v][h]=1; \
	S=surface; S->trigons=0; \
	if (VM==NULL) S->trigon=NULL; else FVM=file=CreateVM(VM,extension)

#define CHECK(h,v)	\
	if (!(back[v][h]=(char)(((back[v][h]&1)?0:2)|((i[y][x]&m)?0:1)))) { \
	    if (h<h1) h1=h; if (h>h2) h2=h; \
	    if (v<v1) v1=v; if (v>v2) v2=v; \
	}

#define TRIANGULATION_YZ(d)	{ \
	h1=v1=N; h2=v2=(-1); \
	for (z=z1; z<=z2; z++) { \
	    i=image[z/8]; m=(char)(1<<z%8); \
	    for (y=Y1; y<=y2; y++) CHECK(y,z) \
	} \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,store); \
	} \
}
	SETUP(Ny,Nz,mx,Fmx,"mx"); for (x=x1; x<=x2; x++) TRIANGULATION_YZ(x  )
	SETUP(Ny,Nz,px,Fpx,"px"); for (x=x2; x>=x1; x--) TRIANGULATION_YZ(x+1)

#define TRIANGULATION_ZX(d)	{ \
	h1=v1=N; h2=v2=(-1); \
	for (z=z1; z<=z2; z++) { \
	    i=image[z/8]; m=(char)(1<<z%8); \
	    for (x=x1; x<=x2; x++) CHECK(z,x) \
	} \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,store); \
	} \
}
	SETUP(Nz,Nx,my,Fmy,"my"); for (y=Y1; y<=y2; y++) TRIANGULATION_ZX(y  )
	SETUP(Nz,Nx,py,Fpy,"py"); for (y=y2; y>=Y1; y--) TRIANGULATION_ZX(y+1)

#define TRIANGULATION_XY(d)	{ \
	h1=v1=N; h2=v2=(-1); \
	i=image[z/8]; m=(char)(1<<z%8); \
	for (y=Y1; y<=y2; y++) \
	for (x=x1; x<=x2; x++) CHECK(x,y) \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,store); \
	} \
}
	SETUP(Nx,Ny,mz,Fmz,"mz"); for (z=z1; z<=z2; z++) TRIANGULATION_XY(z  )
	SETUP(Nx,Ny,pz,Fpz,"pz"); for (z=z2; z>=z1; z--) TRIANGULATION_XY(z+1)

	TermTriangulation(N,back);

	for (z=(Nz-1)/8; z>=0; z--) {
	    for (y=Ny-1; y>=0; y--) free(image[z][y]); free(image[z]);
	}
	free(image);

	if (VM!=NULL) {
	    RestoreVM(Fmx,mx);
	    RestoreVM(Fpx,px);
	    RestoreVM(Fmy,my);
	    RestoreVM(Fpy,py);
	    RestoreVM(Fmz,mz);
	    RestoreVM(Fpz,pz);
	}
}

#define BYTE	unsigned char

static BYTE	**frame,gray;
static int	H,V;

static void	PlotPV(h,v)
int		h,v;
{
	frame[V-v][H+h]=gray;
}

static double	Gamma,bias;

#define SHADE(shade)	(BYTE)(bias+(255.-bias)*pow(shade,1./Gamma)+0.5)

#ifndef	NSS
static void	PlotShade(h,v,shade)
int		h,v;
double		shade;
{
	frame[V-v][H+h]=SHADE(shade);
}
#endif

#define TEXT_LEN	256

static void	Surface_BEV(Nx,Ny,Nz,mx,px,my,py,mz,pz,size,uX,uY,uZ,bg,fg,path)
int		Nx,Ny,Nz,size,bg,fg;
Surface		*mx,*px,*my,*py,*mz,*pz;
double		uX,uY,uZ;
char		*path;
{
	double	Wx,Wy,Wz;
	int	v;
	FILE	*file;
	char	text[TEXT_LEN],
		*fmae="frame memory allocation error.";

	if (size) {
	    Wx=(double)Nx;
	    Wy=(double)Ny;
	    Wz=(double)Nz; uX=
			   uY=
			   uZ=(double)size/sqrt(Wx*Wx+Wy*Wy+Wz*Wz); Wx*=uX;
								    Wy*=uY;
								    Wz*=uZ;
	}
	else {
	    Wx=uX*(double)Nx;
	    Wy=uY*(double)Ny;
	    Wz=uZ*(double)Nz; size=(int)ceil(sqrt(Wx*Wx+Wy*Wy+Wz*Wz));
	}
	if ((frame=ALLOC(BYTE *,size))==NULL) Error(fmae);

	for (v=0; v<size; v++)
	    if ((frame[v]=ALLOC(BYTE,size))==NULL) Error(fmae);

	if (*path=='-')
	    file=stdout;
	else
	    if ((file=fopen(path,"wb"))==NULL) Error("GIF file not open.");

	InitStoreGIF(file,size,size,StoreGIF_Gray,StoreGIF_Gray,StoreGIF_Gray);

	while (fgets(text,TEXT_LEN,stdin)!=NULL)
{
	double	bLo,bLa,sLo,sLa,lo,la,Wh,Wv,side,x,y,z;
	int	cnt,idx,h,n;
	Trigon	*t;

	switch(sscanf(text,"%lf %lf %lf %lf %d",&bLo,&bLa,&sLo,&sLa,&cnt)) {
	    case 2: sLo=sLa=0.; cnt=1; break;
	    case 5: break;
	    default: continue;
	}
	for (idx=0; idx<cnt; idx++) {
	    lo=(bLo+sLo*(double)idx)*M_PI/180.;
	    la=(bLa+sLa*(double)idx)*M_PI/180.;

	    XYZWidth(lo,la,0.,0.,0.,Wx,Wy,Wz,&Wh,&Wv);
	    h=(int)ceil(Wh); H=(size-h)/2;
	    v=(int)ceil(Wv); V=(size+v)/2-1;
	    if (!XYZInit(h,v,lo,la,0.,0.,0.,Wx,Wy,Wz))
		Error("rendering memory allocation error.");

	    for (v=0; v<size; v++)
	    for (h=0; h<size; h++) frame[v][h]=(BYTE)bg;

#ifndef	NSS
#define GRAY(side)
#define PLOTPV		NULL
#else
#define GRAY(side)	gray=SHADE(side);
#define PLOTPV		PlotPV
#endif

	    if ((side=XYZSide(lo,la,0.,0.,0.,0.,1.,0.,0.,0.,1.))!=0.) {
		if (side<0.) {
		    n=mx->trigons; t=mx->trigon; GRAY(-side)
		}
		else {
		    n=px->trigons; t=px->trigon; GRAY( side)
		}
		for (; n; n--, t++) {
		    x=uX*(double)t->d;
		    XYZFill(x,uY*(double)t->hi,uZ*(double)t->vi,
			    x,uY*(double)t->hj,uZ*(double)t->vj,
			    x,uY*(double)t->hk,uZ*(double)t->vk,PLOTPV);
		}
	    }
	    if ((side=XYZSide(lo,la,0.,0.,0.,0.,0.,1.,1.,0.,0.))!=0.) {
		if (side<0.) {
		    n=my->trigons; t=my->trigon; GRAY(-side)
		}
		else {
		    n=py->trigons; t=py->trigon; GRAY( side)
		}
		for (; n; n--, t++) {
		    y=uY*(double)t->d;
		    XYZFill(uX*(double)t->vi,y,uZ*(double)t->hi,
			    uX*(double)t->vj,y,uZ*(double)t->hj,
			    uX*(double)t->vk,y,uZ*(double)t->hk,PLOTPV);
		}
	    }
	    if ((side=XYZSide(lo,la,0.,0.,0.,1.,0.,0.,0.,1.,0.))!=0.) {
		if (side<0.) {
		    n=mz->trigons; t=mz->trigon; GRAY(-side)
		}
		else {
		    n=pz->trigons; t=pz->trigon; GRAY( side)
		}
		for (; n; n--, t++) {
		    z=uZ*(double)t->d;
		    XYZFill(uX*(double)t->hi,uY*(double)t->vi,z,
			    uX*(double)t->hj,uY*(double)t->vj,z,
			    uX*(double)t->hk,uY*(double)t->vk,z,PLOTPV);
		}
	    }
#ifndef	NSS
	    XYZShade(lo,la,PlotShade);
#endif
	    if (fg>=0) {
		gray=(BYTE)fg;
		XYZDraw(0.,0.,0.,Wx,0.,0.,PlotPV);
		XYZDraw(Wx,0.,0.,Wx,Wy,0.,PlotPV);
		XYZDraw(Wx,Wy,0.,0.,Wy,0.,PlotPV);
		XYZDraw(0.,Wy,0.,0.,0.,0.,PlotPV);
		XYZDraw(0.,0.,Wz,Wx,0.,Wz,PlotPV);
		XYZDraw(Wx,0.,Wz,Wx,Wy,Wz,PlotPV);
		XYZDraw(Wx,Wy,Wz,0.,Wy,Wz,PlotPV);
		XYZDraw(0.,Wy,Wz,0.,0.,Wz,PlotPV);
		XYZDraw(0.,0.,0.,0.,0.,Wz,PlotPV);
		XYZDraw(Wx,0.,0.,Wx,0.,Wz,PlotPV);
		XYZDraw(Wx,Wy,0.,Wx,Wy,Wz,PlotPV);
		XYZDraw(0.,Wy,0.,0.,Wy,Wz,PlotPV);
	    }
	    XYZTerm();

	    StoreGIF(file,0,0,size,size,NULL,NULL,NULL,frame);
	}
}
	(void)TermStoreGIF(file); if (file!=stdout) (void)fclose(file);

	for (v=size-1; v>=0; v--) free(frame[v]); free(frame);
}

#define PV	(int)((Cell)(~0))

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**path,solid[PV+1];
	int	Nx,Ny,Nz,size,bg,fg;
	double	uX,uY,uZ;
	Surface	mx,px,my,py,mz,pz;

	if (argc<10 || argc>13)
	Error("usage : si_s_bev directory nameFile rangeList {VM}\n"
	      "                 size|(unitX unitY unitZ) gamma bias bg fg GIF");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],PV,solid);

	if (argc==12 || argc==13) {
	    if ((uX=atof(argv[argc-8]))<=0. ||
		(uY=atof(argv[argc-7]))<=0. ||
		(uZ=atof(argv[argc-6]))<=0.) Error("bad unit.");

	    size=0;
	}
	else
	    if ((size=atoi(argv[argc-6]))<=0) Error("bad size.");

	if ((Gamma=atof(argv[argc-5]))<=0.) Error("bad gamma.");
	if ((bias=atof(argv[argc-4]))<0. || bias>255.) Error("bad bias.");

	if ((bg=atoi(argv[argc-3]))<0 || bg>255) Error("bad background.");
	if ((fg=atoi(argv[argc-2]))>255) Error("bad foreground.");

	SI_Surface(path,Nx,Ny,Nz,solid,
		   (argc==11 || argc==13)?argv[4]:NULL,
		   &mx,&px,&my,&py,&mz,&pz);

	Surface_BEV(Nx,Ny,Nz,&mx,&px,&my,&py,&mz,&pz,
		    size,uX,uY,uZ,bg,fg,argv[argc-1]);

	return 0;
}
