
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "t_s.h"

extern void	Error(),RangeList();

static char	*STL="facet normal %d %d %d\n"
		     "outer loop\nvertex %d %d %d\n"
				 "vertex %d %d %d\n"
				 "vertex %d %d %d\nendloop\nendfacet\n";
static int	D;

#define STORE(function,nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk)	\
	static void	function(hi,vi,hj,vj,hk,vk) \
	int		hi,vi,hj,vj,hk,vk; \
	{ \
		(void)printf(STL,nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk); \
	}

STORE(Mx,-1, 0, 0,D ,hi,vi,D ,hj,vj,D ,hk,vk)
STORE(Px, 1, 0, 0,D ,hi,vi,D ,hj,vj,D ,hk,vk)
STORE(My, 0,-1, 0,vi,D ,hi,vj,D ,hj,vk,D ,hk)
STORE(Py, 0, 1, 0,vi,D ,hi,vj,D ,hj,vk,D ,hk)
STORE(Mz, 0, 0,-1,hi,vi,D ,hj,vj,D ,hk,vk,D )
STORE(Pz, 0, 0, 1,hi,vi,D ,hj,vj,D ,hk,vk,D )

#define PV		(int)((Cell)(~0))
#define ALLOC(type,noe)	(type *)malloc((size_t)(noe)*sizeof(type))

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**path,solid[PV+1],***image,**i,m,**back,
		*mae="memory allocation error.";
	int	Nx,Ny,Nz,x1,y1,z1,x2,y2,z2,z,y,x,nx,ny,N,v,h,h1,v1,h2,v2;
	Cell	**cell;

	if (argc!=4) Error("usage : si_s_stl directory nameFile rangeList");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	RangeList(argv[3],PV,solid);

	if ((image=ALLOC(char **,(Nz+7)/8))==NULL) Error(mae);

	x1=Nx; y1=Ny; z1=Nz; x2=y2=z2=(-1);
	for (z=0; z<Nz; z++) {
	    if (z%8)
		m<<=1;
	    else {
		if ((image[z/8]=i=ALLOC(char *,Ny))==NULL) Error(mae);

		for (y=0; y<Ny; y++) {
		    if ((i[y]=ALLOC(char,Nx))==NULL) Error(mae);

		    for (x=0; x<Nx; x++) i[y][x]=0;
		}
		m=1;
	    }
	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL);

	    for (y=0; y<ny && y<Ny; y++)
	    for (x=0; x<nx && x<Nx; x++)
		if (solid[cell[y][x]]) {
		    i[y][x]|=m; if (x<x1) x1=x; if (x>x2) x2=x;
				if (y<y1) y1=y; if (y>y2) y2=y;
				if (z<z1) z1=z; if (z>z2) z2=z;
		}

	    for (y=ny-1; y>=0; y--) free(cell[y]); free(cell);
	}
	if (x1>x2 || y1>y2 || z1>z2) Error("no solid object.");

	N=(Nx>Ny)?(Nx>Nz)?Nx:Nz:(Ny>Nz)?Ny:Nz;

	InitTriangulation(N,N,&back); (void)puts("solid");

#define CLEAR(Nh,Nv)	for (v=0; v<Nv; v++) for (h=0; h<Nh; h++) back[v][h]=1

#define CHECK(h,v)	\
	if (!(back[v][h]=(char)(((back[v][h]&1)?0:2)|((i[y][x]&m)?0:1)))) { \
	    if (h<h1) h1=h; if (h>h2) h2=h; \
	    if (v<v1) v1=v; if (v>v2) v2=v; \
	}

#define TRIANGULATION_YZ(d,store)	{ \
	h1=v1=N; h2=v2=(-1); \
	for (z=z1; z<=z2; z++) { \
	    i=image[z/8]; m=(char)(1<<z%8); \
	    for (y=y1; y<=y2; y++) CHECK(y,z) \
	} \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,store); \
	} \
}
	CLEAR(Ny,Nz); for (x=x1; x<=x2; x++) TRIANGULATION_YZ(x  ,Mx)
	CLEAR(Ny,Nz); for (x=x2; x>=x1; x--) TRIANGULATION_YZ(x+1,Px)

#define TRIANGULATION_ZX(d,store)	{ \
	h1=v1=N; h2=v2=(-1); \
	for (z=z1; z<=z2; z++) { \
	    i=image[z/8]; m=(char)(1<<z%8); \
	    for (x=x1; x<=x2; x++) CHECK(z,x) \
	} \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,store); \
	} \
}
	CLEAR(Nz,Nx); for (y=y1; y<=y2; y++) TRIANGULATION_ZX(y  ,My)
	CLEAR(Nz,Nx); for (y=y2; y>=y1; y--) TRIANGULATION_ZX(y+1,Py)

#define TRIANGULATION_XY(d,store)	{ \
	h1=v1=N; h2=v2=(-1); \
	i=image[z/8]; m=(char)(1<<z%8); \
	for (y=y1; y<=y2; y++) \
	for (x=x1; x<=x2; x++) CHECK(x,y) \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,store); \
	} \
}
	CLEAR(Nx,Ny); for (z=z1; z<=z2; z++) TRIANGULATION_XY(z  ,Mz)
	CLEAR(Nx,Ny); for (z=z2; z>=z1; z--) TRIANGULATION_XY(z+1,Pz)

	(void)puts("endsolid"); TermTriangulation(N,back); return 0;
}
