
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "t_s.h"
#include "xyz.h"
#include "sgif.h"

extern void	Error();

#define BYTE	unsigned char

static BYTE	**frame,gray;
static int	H,V;

static void	PlotPV(h,v)
int		h,v;
{
	frame[V-v][H+h]=gray;
}

#ifndef	NSS
#define PLOTPV	NULL
#else
#define PLOTPV	PlotPV
#endif

static double	uX,uY,uZ,D;

static void	PlotX(hi,vi,hj,vj,hk,vk)
int		hi,vi,hj,vj,hk,vk;
{
	XYZFill(D,uY*(double)hi,uZ*(double)vi,
		D,uY*(double)hj,uZ*(double)vj,
		D,uY*(double)hk,uZ*(double)vk,PLOTPV);
}

static void	PlotY(hi,vi,hj,vj,hk,vk)
int		hi,vi,hj,vj,hk,vk;
{
	XYZFill(uX*(double)vi,D,uZ*(double)hi,
		uX*(double)vj,D,uZ*(double)hj,
		uX*(double)vk,D,uZ*(double)hk,PLOTPV);
}

static void	PlotZ(hi,vi,hj,vj,hk,vk)
int		hi,vi,hj,vj,hk,vk;
{
	XYZFill(uX*(double)hi,uY*(double)vi,D,
		uX*(double)hj,uY*(double)vj,D,
		uX*(double)hk,uY*(double)vk,D,PLOTPV);
}

static double	Gamma,bias;

#define SHADE(shade)	(BYTE)(bias+(255.-bias)*pow(shade,1./Gamma)+0.5)

#ifndef	NSS
static void	PlotShade(h,v,shade)
int		h,v;
double		shade;
{
	frame[V-v][H+h]=SHADE(shade);
}
#endif

#define TEXT_LEN	256
#define ALLOC(type,noe)	(type *)malloc((size_t)(noe)*sizeof(type))
#define PV		(int)((Cell)(~0))

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**path,**back,text[TEXT_LEN],
		*imae="image memory allocation error.",
		*fmae="frame memory allocation error.";
	int	Nx,Ny,Nz,size,bg,fg,z,nx,ny,y,x,N,v;
	Cell	***image,**cell,**slice;
	double	Wx,Wy,Wz;
	FILE	*file;

	if (argc!=9 && argc!=11)
	Error("usage : si_t_bev directory nameFile\n"
	      "                 size|(unitX unitY unitZ) gamma bias bg fg GIF");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	if (argc==11) {
	    if ((uX=atof(argv[3]))<=0. ||
		(uY=atof(argv[4]))<=0. ||
		(uZ=atof(argv[5]))<=0.) Error("bad unit.");

	    size=0;
	}
	else
	    if ((size=atoi(argv[3]))<=0) Error("bad size.");

	if ((Gamma=atof(argv[argc-5]))<=0.) Error("bad gamma.");
	if ((bias=atof(argv[argc-4]))<0. || bias>255.) Error("bad bias.");

	if ((bg=atoi(argv[argc-3]))<0 || bg>255) Error("bad background.");
	if ((fg=atoi(argv[argc-2]))>255) Error("bad foreground.");

	if ((image=ALLOC(Cell **,Nz))==NULL) Error(imae);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL); free(path[z]);

	    if (nx>=Nx && ny>=Ny)
		image[z]=cell;
	    else {
		if ((image[z]=slice=ALLOC(Cell *,Ny))==NULL) Error(imae);

		for (y=0; y<Ny; y++) {
		    if ((slice[y]=ALLOC(Cell,Nx))==NULL) Error(imae);

		    for (x=0; x<Nx; x++)
			slice[y][x]=(x<nx && y<ny)?cell[y][x]:0;
		}
		for (y=ny-1; y>=0; y--) free(cell[y]); free(cell);
	    }
	}
	free(path);

	N=(Nx>Ny)?(Nx>Nz)?Nx:Nz:(Ny>Nz)?Ny:Nz;

	InitTriangulation(N,N,&back);

	if (size) {
	    Wx=(double)Nx;
	    Wy=(double)Ny;
	    Wz=(double)Nz; uX=
			   uY=
			   uZ=(double)size/sqrt(Wx*Wx+Wy*Wy+Wz*Wz); Wx*=uX;
								    Wy*=uY;
								    Wz*=uZ;
	}
	else {
	    Wx=uX*(double)Nx;
	    Wy=uY*(double)Ny;
	    Wz=uZ*(double)Nz; size=(int)ceil(sqrt(Wx*Wx+Wy*Wy+Wz*Wz));
	}
	if ((frame=ALLOC(BYTE *,size))==NULL) Error(fmae);

	for (v=0; v<size; v++)
	    if ((frame[v]=ALLOC(BYTE,size))==NULL) Error(fmae);

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	if ((file=fopen(argv[argc-1],"wb"))==NULL) Error("GIF file not open.");

	InitStoreGIF(file,size,size,StoreGIF_Gray,StoreGIF_Gray,StoreGIF_Gray);

	while (fgets(text,TEXT_LEN,stdin)!=NULL)
{
	int	bC,sC,cnt,idx,C,pv,h,x1,Y1,z1,x2,y2,z2,h1,v1,h2,v2;
	double	bLo,bLa,sLo,sLa,lo,la,Wh,Wv,side;
	char	solid[PV+1];

	switch(sscanf(text,"%d %lf %lf %d %lf %lf %d",
			   &bC,&bLo,&bLa,&sC,&sLo,&sLa,&cnt)) {
	    case 3: sC=0; sLo=sLa=0.; cnt=1; break;
	    case 7: break;
	    default: continue;
	}
	for (idx=0; idx<cnt; idx++) {
	    if ((C=bC+sC*idx)>=0)
		for (pv=0; pv<=PV; pv++) solid[pv]=pv>=C;
	    else
		for (pv=0; pv<=PV; pv++) solid[pv]=pv<-C;

	    lo=(bLo+sLo*(double)idx)*M_PI/180.;
	    la=(bLa+sLa*(double)idx)*M_PI/180.;

	    XYZWidth(lo,la,0.,0.,0.,Wx,Wy,Wz,&Wh,&Wv);
	    h=(int)ceil(Wh); H=(size-h)/2;
	    v=(int)ceil(Wv); V=(size+v)/2-1;
	    if (!XYZInit(h,v,lo,la,0.,0.,0.,Wx,Wy,Wz))
		Error("rendering memory allocation error.");

	    for (v=0; v<size; v++)
	    for (h=0; h<size; h++) frame[v][h]=(BYTE)bg;

	    x1=Nx; Y1=Ny; z1=Nz; x2=y2=z2=(-1);
	    for (z=0; z<Nz; z++)
	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++)
		if (solid[image[z][y][x]]) {
		    if (x<x1) x1=x; if (x>x2) x2=x;
		    if (y<Y1) Y1=y; if (y>y2) y2=y;
		    if (z<z1) z1=z; if (z>z2) z2=z;
		}

	    if (x1<=x2 && Y1<=y2 && z1<=z2) {

#define CLEAR(Nh,Nv)	for (v=0; v<Nv; v++) for (h=0; h<Nh; h++) back[v][h]=1

#ifndef	NSS
#define GRAY(shade)
#else
#define GRAY(shade)	gray=SHADE(shade);
#endif

#define	TRIANGULATION(H,H1,H2,V,V1,V2,d,plot)	{ \
	h1=v1=N; h2=v2=(-1); \
	for (V=V1; V<=V2; V++) \
	for (H=H1; H<=H2; H++) \
	    if (!(back[V][H]=(char)(((back[V][H]&1)?0:2)| \
				    ((solid[image[z][y][x]])?0:1)))) { \
		if (H<h1) h1=H; if (H>h2) h2=H; \
		if (V<v1) v1=V; if (V>v2) v2=V; \
	    } \
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,plot); \
	} \
}
		if ((side=XYZSide(lo,la,0.,0.,0.,0.,1.,0.,0.,0.,1.))!=0.) {
		    CLEAR(Ny,Nz);
		    if (side<0.) {
			GRAY(-side)
			for (x=x1; x<=x2; x++)
			TRIANGULATION(y,Y1,y2,z,z1,z2,uX*(double) x   ,PlotX)
		    }
		    else {
			GRAY( side)
			for (x=x2; x>=x1; x--)
			TRIANGULATION(y,Y1,y2,z,z1,z2,uX*(double)(x+1),PlotX)
		    }
		}
		if ((side=XYZSide(lo,la,0.,0.,0.,0.,0.,1.,1.,0.,0.))!=0.) {
		    CLEAR(Nz,Nx);
		    if (side<0.) {
			GRAY(-side)
			for (y=Y1; y<=y2; y++)
			TRIANGULATION(z,z1,z2,x,x1,x2,uY*(double) y   ,PlotY)
		    }
		    else {
			GRAY( side)
			for (y=y2; y>=Y1; y--)
			TRIANGULATION(z,z1,z2,x,x1,x2,uY*(double)(y+1),PlotY)
		    }
		}
		if ((side=XYZSide(lo,la,0.,0.,0.,1.,0.,0.,0.,1.,0.))!=0.) {
		    CLEAR(Nx,Ny);
		    if (side<0.) {
			GRAY(-side)
			for (z=z1; z<=z2; z++)
			TRIANGULATION(x,x1,x2,y,Y1,y2,uZ*(double) z   ,PlotZ)
		    }
		    else {
			GRAY( side)
			for (z=z2; z>=z1; z--)
			TRIANGULATION(x,x1,x2,y,Y1,y2,uZ*(double)(z+1),PlotZ)
		    }
		}
#ifndef	NSS
		XYZShade(lo,la,PlotShade);
#endif
	    }
	    if (fg>=0) {
		gray=(BYTE)fg;
		XYZDraw(0.,0.,0.,Wx,0.,0.,PlotPV);
		XYZDraw(Wx,0.,0.,Wx,Wy,0.,PlotPV);
		XYZDraw(Wx,Wy,0.,0.,Wy,0.,PlotPV);
		XYZDraw(0.,Wy,0.,0.,0.,0.,PlotPV);
		XYZDraw(0.,0.,Wz,Wx,0.,Wz,PlotPV);
		XYZDraw(Wx,0.,Wz,Wx,Wy,Wz,PlotPV);
		XYZDraw(Wx,Wy,Wz,0.,Wy,Wz,PlotPV);
		XYZDraw(0.,Wy,Wz,0.,0.,Wz,PlotPV);
		XYZDraw(0.,0.,0.,0.,0.,Wz,PlotPV);
		XYZDraw(Wx,0.,0.,Wx,0.,Wz,PlotPV);
		XYZDraw(Wx,Wy,0.,Wx,Wy,Wz,PlotPV);
		XYZDraw(0.,Wy,0.,0.,Wy,Wz,PlotPV);
	    }
	    XYZTerm();

	    StoreGIF(file,0,0,size,size,NULL,NULL,NULL,frame);
	}
}
	(void)TermStoreGIF(file); if (file!=stdout) (void)fclose(file);

	return 0;
}
