
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "t_m.h"
#include "xyz.h"
#include "sgif.h"

#ifndef	XYZ
#define XYZ	3
#endif
#if	XYZ<0 || XYZ>3
#undef	XYZ
#define XYZ	3
#endif

extern void	Error(),ColorFile();

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(size_t)(noe))

static char	*imae="image memory allocation error.";

static Cell	***ReadSI(path,Nx,Ny,Nz)
char		**path;
int		Nx,Ny,Nz;
{
	Cell	***image,**cell,**slice;
	int	z,nx,ny,y,x;

	if ((image=ALLOC(Cell **,Nz))==NULL) Error(imae);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL); free(path[z]);

	    if (nx>=Nx && ny>=Ny)
		image[z]=cell;
	    else {
		if ((image[z]=slice=ALLOC(Cell *,Ny))==NULL) Error(imae);

		for (y=0; y<Ny; y++) {
		    if ((slice[y]=ALLOC(Cell,Nx))==NULL) Error(imae);

		    for (x=0; x<Nx; x++)
			slice[y][x]=(x<nx && y<ny)?cell[y][x]:0;
		}
		for (y=ny-1; y>=0; y--) free(cell[y]); free(cell);
	    }
	}
	free(path); return image;
}

static Cell	***ZeroSI(Nx,Ny,Nz)
int		Nx,Ny,Nz;
{
	Cell	***image;
	int	x,y,z;

	if ((  image=ALLOC(Cell **,Nz))==NULL ||
	    ( *image=ALLOC(Cell * ,Ny))==NULL ||
	    (**image=ALLOC(Cell   ,Nx))==NULL) Error(imae);

	***image=0; for (x=1; x<Nx; x++) (**image)[x]=(***image);
		    for (y=1; y<Ny; y++)  (*image)[y]=( **image);
		    for (z=1; z<Nz; z++)    image [z]=(  *image); return image;
}

#define BYTE	unsigned char

static void	ScanSI(Nx,Ny,Nz,image,solid,R,G,B,x1,Y1,z1,x2,y2,z2,gray)
int		Nx,Ny,Nz,*x1,*Y1,*z1,*x2,*y2,*z2;
Cell		***image;
char		*solid;
BYTE		*R,*G,*B,**gray;
{
	int	z,y,x,pv;

	for (z=0; z<Nz; z++)
	for (y=0; y<Ny; y++)
	for (x=0; x<Nx; x++)
	    if (solid[pv=image[z][y][x]]) {
		if (x<*x1) *x1=x; if (x>*x2) *x2=x;
		if (y<*Y1) *Y1=y; if (y>*y2) *y2=y;
		if (z<*z1) *z1=z; if (z>*z2) *z2=z;

		if (*gray!=NULL && (R[pv]!=G[pv] || G[pv]!=B[pv])) *gray=NULL;
	    }
}

static BYTE	**r,**g,**b,pvR,pvG,pvB;
static int	H,V;

static void	PlotPV(h,v)
int		h,v;
{
	h+=H; v=V-v; r[v][h]=pvR; g[v][h]=pvG; b[v][h]=pvB;
}

#ifndef	NSS
#define	COLOR_RGB()	pvR=(BYTE)(c    &0xff); \
			pvG=(BYTE)(c>>8 &0xff); \
			pvB=(BYTE)(c>>16&0xff)
#else
static double	I;

#define COLOR_RGB()	pvR=(BYTE)(I*(double)(c    &0xff)+0.5); \
			pvG=(BYTE)(I*(double)(c>>8 &0xff)+0.5); \
			pvB=(BYTE)(I*(double)(c>>16&0xff)+0.5)
#endif

static double	uX,uY,uZ,D;

static void	PlotX(hi,vi,hj,vj,hk,vk,c)
int		hi,vi,hj,vj,hk,vk;
long		c;
{
	COLOR_RGB(); XYZFill(D,uY*(double)hi,uZ*(double)vi,
			     D,uY*(double)hj,uZ*(double)vj,
			     D,uY*(double)hk,uZ*(double)vk,PlotPV);
}

static void	PlotY(hi,vi,hj,vj,hk,vk,c)
int		hi,vi,hj,vj,hk,vk;
long		c;
{
	COLOR_RGB(); XYZFill(uX*(double)vi,D,uZ*(double)hi,
			     uX*(double)vj,D,uZ*(double)hj,
			     uX*(double)vk,D,uZ*(double)hk,PlotPV);
}
static void	PlotZ(hi,vi,hj,vj,hk,vk,c)
int		hi,vi,hj,vj,hk,vk;
long		c;
{
	COLOR_RGB(); XYZFill(uX*(double)hi,uY*(double)vi,D,
			     uX*(double)hj,uY*(double)vj,D,
			     uX*(double)hk,uY*(double)vk,D,PlotPV);
}

static double	Gamma,bias;

#define SHADE(shade)	(bias+(255.-bias)*pow(shade,1./Gamma))/255.

#ifndef	NSS
static void	PlotShade(h,v,shade)
int		h,v;
double		shade;
{
	double	I=SHADE(shade);

	h+=H; v=V-v; r[v][h]=(BYTE)(I*(double)r[v][h]+0.5);
		     g[v][h]=(BYTE)(I*(double)g[v][h]+0.5);
		     b[v][h]=(BYTE)(I*(double)b[v][h]+0.5);
}
#endif

#define PV		(int)((Cell)(~0))
#define TEXT_LEN	256

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	size,bgR,bgG,bgB,fgR,fgG,fgB,Nx,Ny,Nz,x1,Y1,z1,x2,y2,z2,N,v;
	BYTE	R1[PV+1],G1[PV+1],B1[PV+1],R2[PV+1],G2[PV+1],B2[PV+1],
		*gray=StoreGIF_Gray;
	Cell	***image1,***image2;
	char	solid1[PV+1],solid2[PV+1],**back,text[TEXT_LEN],
		*fmae="frame memory allocation error.";
	long	**fore;
	double	Wx,Wy,Wz;
	FILE	*file;

	if ((argc!=17 && argc!=19) || (argv[1][0]=='-' && argv[4][0]=='-'))
	    Error("usage : "
#if	XYZ==1
		  "si_x_bev "
#endif
#if	XYZ==2
		  "si_y_bev "
#endif
#if	XYZ==3
		  "si_z_bev "
#endif
		  "directory_1 nameFile_1 colorFile_1\n\t\t "
		  "directory_2 nameFile_2 colorFile_2\n\t\t "
		  "size|(unitX unitY unitZ) gamma bias "
		  "bgR bgG bgB fgR fgG fgB GIF");

	if (argc==19) {
	    if ((uX=atof(argv[7]))<=0. ||
		(uY=atof(argv[8]))<=0. ||
		(uZ=atof(argv[9]))<=0.) Error("bad unit.");

	    size=0;
	}
	else
	    if ((size=atoi(argv[7]))<=0) Error("bad size.");

	if ((Gamma=atof(argv[argc-9]))<=0.) Error("bad gamma.");
	if ((bias=atof(argv[argc-8]))<0. || bias>255.) Error("bad bias.");

	if ((bgR=atoi(argv[argc-7]))<0 || bgR>255 ||
	    (bgG=atoi(argv[argc-6]))<0 || bgG>255 ||
	    (bgB=atoi(argv[argc-5]))<0 || bgB>255) Error("bad background.");
	if ((fgR=atoi(argv[argc-4]))>255 ||
	    (fgG=atoi(argv[argc-3]))>255 ||
	    (fgB=atoi(argv[argc-2]))>255) Error("bad foreground.");

	if (fgR<0 || fgG<0 || fgB<0) fgR=fgG=fgB=(-1);

	if (bgR!=bgG || bgG!=bgB || fgR!=fgG || fgG!=fgB) gray=NULL;

	ColorFile(argv[3],solid1,R1,G1,B1);
	ColorFile(argv[6],solid2,R2,G2,B2);
{
	char	**path;
	int	nx,ny,nz;

	if (argv[1][0]!='-') {
	    path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);
	    image1=ReadSI(path,Nx,Ny,Nz);
	    if (argv[4][0]!='-') {
		if (strcmp(argv[1],argv[4]) || strcmp(argv[2],argv[5])) {
		    path=CheckSliceImages(argv[4],argv[5],&nx,&ny,&nz,NULL);
		    if (nx!=Nx || ny!=Ny || nz!=Nz) Error("bad pixel number.");

		    image2=ReadSI(path,Nx,Ny,Nz);
		}
		else
		    image2=image1;

		x1=Nx; Y1=Ny; z1=Nz; x2=y2=z2=(-1);
		ScanSI(Nx,Ny,Nz,image1,solid1,R1,G1,B1,
		       &x1,&Y1,&z1,&x2,&y2,&z2,&gray);
		ScanSI(Nx,Ny,Nz,image2,solid2,R2,G2,B2,
		       &x1,&Y1,&z1,&x2,&y2,&z2,&gray);
	    }
	    else {
		image2=ZeroSI(Nx,Ny,Nz);
		if (*solid2) {
		    x1=Y1=z1=0; x2=Nx-1; y2=Ny-1; z2=Nz-1;
		    if (gray!=NULL)
			if (*R2!=(*G2) || *G2!=(*B2))
			    gray=NULL;
			else
			    ScanSI(Nx,Ny,Nz,image1,solid1,R1,G1,B1,
				   &x1,&Y1,&z1,&x2,&y2,&z2,&gray);
		}
		else {
		    x1=Nx; Y1=Ny; z1=Nz; x2=y2=z2=(-1);
		    ScanSI(Nx,Ny,Nz,image1,solid1,R1,G1,B1,
			   &x1,&Y1,&z1,&x2,&y2,&z2,&gray);
		}
	    }
	}
	else {
	    path=CheckSliceImages(argv[4],argv[5],&Nx,&Ny,&Nz,NULL);
	    image2=ReadSI(path,Nx,Ny,Nz);
	    image1=ZeroSI(Nx,Ny,Nz);
	    if (*solid1) {
		x1=Y1=z1=0; x2=Nx-1; y2=Ny-1; z2=Nz-1;
		if (gray!=NULL)
		    if (*R1!=(*G1) || *G1!=(*B1))
			gray=NULL;
		    else
			ScanSI(Nx,Ny,Nz,image2,solid2,R2,G2,B2,
			       &x1,&Y1,&z1,&x2,&y2,&z2,&gray);
	    }
	    else {
		x1=Nx; Y1=Ny; z1=Nz; x2=y2=z2=(-1);
		ScanSI(Nx,Ny,Nz,image2,solid2,R2,G2,B2,
		       &x1,&Y1,&z1,&x2,&y2,&z2,&gray);
	    }
	}
}
	if (x1>x2 || Y1>y2 || z1>z2) Error("no solid object.");

	N=(Nx>Ny)?(Nx>Nz)?Nx:Nz:(Ny>Nz)?Ny:Nz;

	InitTriangulation(N,N,&back,&fore);

	if (size) {
	    Wx=(double)Nx;
	    Wy=(double)Ny;
	    Wz=(double)Nz; uX=
			   uY=
			   uZ=(double)size/sqrt(Wx*Wx+Wy*Wy+Wz*Wz); Wx*=uX;
								    Wy*=uY;
								    Wz*=uZ;
	}
	else {
	    Wx=uX*(double)Nx;
	    Wy=uY*(double)Ny;
	    Wz=uZ*(double)Nz; size=(int)ceil(sqrt(Wx*Wx+Wy*Wy+Wz*Wz));
	}
	if (gray==NULL) {
	    if ((r=ALLOC(BYTE *,size))==NULL ||
		(g=ALLOC(BYTE *,size))==NULL ||
		(b=ALLOC(BYTE *,size))==NULL) Error(fmae);

	    for (v=0; v<size; v++)
		if ((r[v]=ALLOC(BYTE,size))==NULL ||
		    (g[v]=ALLOC(BYTE,size))==NULL ||
		    (b[v]=ALLOC(BYTE,size))==NULL) Error(fmae);
	}
	else {
	    if ((r=g=b=ALLOC(BYTE *,size))==NULL) Error(fmae);

	    for (v=0; v<size; v++)
		if ((r[v]=ALLOC(BYTE,size))==NULL) Error(fmae);
	}
	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	if ((file=fopen(argv[argc-1],"wb"))==NULL) Error("GIF file not open.");

	InitStoreGIF(file,size,size,gray,gray,gray);

	while (fgets(text,TEXT_LEN,stdin)!=NULL)
{
	int	bC,sC,cnt,idx,C,h,x,y,z,h1,v1,h2,v2;
	double	bLo,bLa,sLo,sLa,lo,la,Wh,Wv,side;
	Cell	***image,pv;
	char	*solid;
	BYTE	*R,*G,*B;

	switch(sscanf(text,"%d %lf %lf %d %lf %lf %d",
			   &bC,&bLo,&bLa,&sC,&sLo,&sLa,&cnt)) {
	    case 3: sC=0; sLo=sLa=0.; cnt=1; break;
	    case 7: break;
	    default: continue;
	}
	for (idx=0; idx<cnt; idx++) {
	    C=bC+sC*idx; lo=(bLo+sLo*(double)idx)*M_PI/180.;
			 la=(bLa+sLa*(double)idx)*M_PI/180.;

	    XYZWidth(lo,la,0.,0.,0.,Wx,Wy,Wz,&Wh,&Wv);
	    h=(int)ceil(Wh); H=(size-h)/2;
	    v=(int)ceil(Wv); V=(size+v)/2-1;
	    if (!XYZInit(h,v,lo,la,0.,0.,0.,Wx,Wy,Wz))
		Error("rendering memory allocation error.");

	    for (v=0; v<size; v++)
	    for (h=0; h<size; h++) {
		r[v][h]=(BYTE)bgR; g[v][h]=(BYTE)bgG; b[v][h]=(BYTE)bgB;
	    }

#define CLEAR(Nh,Nv)	for (v=0; v<Nv; v++) for (h=0; h<Nh; h++) back[v][h]=1

#ifndef	NSS
#define	COLOR_I(shade)
#else
#define COLOR_I(shade)	I=SHADE(shade);
#endif

#define SELECT(xyz)	if (xyz<C) { \
			    image=image1; solid=solid1; R=R1; G=G1; B=B1; \
			} \
			else { \
			    image=image2; solid=solid2; R=R2; G=G2; B=B2; \
			}

#if	XYZ==1
#define SELECT_X()	SELECT(x)
#else
#define SELECT_X()
#endif
#if	XYZ==2
#define SELECT_Y()	SELECT(y)
#else
#define SELECT_Y()
#endif
#if	XYZ==3
#define SELECT_Z()	SELECT(z)
#else
#define SELECT_Z()
#endif

#define CHECK(H,V)	\
	if (!(back[V][H]=(char)(((back[V][H]&1)?0:2)| \
				((solid[pv=image[z][y][x]])?0:1)))) { \
	    fore[V][H]=(long)R[pv]|(long)G[pv]<<8|(long)B[pv]<<16; \
	    if (H<h1) h1=H; if (H>h2) h2=H; \
	    if (V<v1) v1=V; if (V>v2) v2=V; \
	}

#define TRIANGULATION(d,plot)	\
	if (h1<=h2 && v1<=v2) { \
	    D=d; Triangulation(h1,v1,h2,v2,back,fore,plot); \
	}

	    if ((side=XYZSide(lo,la,0.,0.,0.,0.,1.,0.,0.,0.,1.))!=0.) {
		CLEAR(Ny,Nz);
		if (side<0.) {
		    COLOR_I(-side)
		    for (x=x1; x<=x2; x++) { SELECT_X()
			h1=v1=N; h2=v2=(-1);
			for (z=z1; z<=z2; z++) { SELECT_Z()
			for (y=Y1; y<=y2; y++) { SELECT_Y()
			    CHECK(y,z)
			}}
			TRIANGULATION(uX*(double) x   ,PlotX)
		    }
		}
		else {
		    COLOR_I( side)
		    for (x=x2; x>=x1; x--) { SELECT_X()
			h1=v1=N; h2=v2=(-1);
			for (z=z1; z<=z2; z++) { SELECT_Z()
			for (y=Y1; y<=y2; y++) { SELECT_Y()
			    CHECK(y,z)
			}}
			TRIANGULATION(uX*(double)(x+1),PlotX)
		    }
		}
	    }
	    if ((side=XYZSide(lo,la,0.,0.,0.,0.,0.,1.,1.,0.,0.))!=0.) {
		CLEAR(Nz,Nx);
		if (side<0.) {
		    COLOR_I(-side)
		    for (y=Y1; y<=y2; y++) { SELECT_Y()
			h1=v1=N; h2=v2=(-1);
			for (x=x1; x<=x2; x++) { SELECT_X()
			for (z=z1; z<=z2; z++) { SELECT_Z()
			    CHECK(z,x)
			}}
			TRIANGULATION(uY*(double) y   ,PlotY)
		    }
		}
		else {
		    COLOR_I( side)
		    for (y=y2; y>=Y1; y--) { SELECT_Y()
			h1=v1=N; h2=v2=(-1);
			for (x=x1; x<=x2; x++) { SELECT_X()
			for (z=z1; z<=z2; z++) { SELECT_Z()
			    CHECK(z,x)
			}}
			TRIANGULATION(uY*(double)(y+1),PlotY)
		    }
		}
	    }
	    if ((side=XYZSide(lo,la,0.,0.,0.,1.,0.,0.,0.,1.,0.))!=0.) {
		CLEAR(Nx,Ny);
		if (side<0.) {
		    COLOR_I(-side)
		    for (z=z1; z<=z2; z++) { SELECT_Z()
			h1=v1=N; h2=v2=(-1);
			for (y=Y1; y<=y2; y++) { SELECT_Y()
			for (x=x1; x<=x2; x++) { SELECT_X()
			    CHECK(x,y)
			}}
			TRIANGULATION(uZ*(double) z   ,PlotZ)
		    }
		}
		else {
		    COLOR_I( side)
		    for (z=z2; z>=z1; z--) { SELECT_Z()
			h1=v1=N; h2=v2=(-1);
			for (y=Y1; y<=y2; y++) { SELECT_Y()
			for (x=x1; x<=x2; x++) { SELECT_X()
			    CHECK(x,y)
			}}
			TRIANGULATION(uZ*(double)(z+1),PlotZ)
		    }
		}
	    }
#ifndef	NSS
	    XYZShade(lo,la,PlotShade);
#endif
	    if (fgR>=0 && fgG>=0 && fgB>=0) {
		pvR=(BYTE)fgR; pvG=(BYTE)fgG; pvB=(BYTE)fgB;
		XYZDraw(0.,0.,0.,Wx,0.,0.,PlotPV);
		XYZDraw(Wx,0.,0.,Wx,Wy,0.,PlotPV);
		XYZDraw(Wx,Wy,0.,0.,Wy,0.,PlotPV);
		XYZDraw(0.,Wy,0.,0.,0.,0.,PlotPV);
		XYZDraw(0.,0.,Wz,Wx,0.,Wz,PlotPV);
		XYZDraw(Wx,0.,Wz,Wx,Wy,Wz,PlotPV);
		XYZDraw(Wx,Wy,Wz,0.,Wy,Wz,PlotPV);
		XYZDraw(0.,Wy,Wz,0.,0.,Wz,PlotPV);
		XYZDraw(0.,0.,0.,0.,0.,Wz,PlotPV);
		XYZDraw(Wx,0.,0.,Wx,0.,Wz,PlotPV);
		XYZDraw(Wx,Wy,0.,Wx,Wy,Wz,PlotPV);
		XYZDraw(0.,Wy,0.,0.,Wy,Wz,PlotPV);
#if	XYZ==1
		if (C>0 && C<Nx)
{
	double	Cx=uX*(double)C;

	XYZDraw(Cx,0.,0.,Cx,Wy,0.,PlotPV);
	XYZDraw(Cx,Wy,0.,Cx,Wy,Wz,PlotPV);
	XYZDraw(Cx,Wy,Wz,Cx,0.,Wz,PlotPV);
	XYZDraw(Cx,0.,Wz,Cx,0.,0.,PlotPV);
}
#endif
#if	XYZ==2
		if (C>0 && C<Ny)
{
	double	Cy=uY*(double)C;

	XYZDraw(0.,Cy,0.,0.,Cy,Wz,PlotPV);
	XYZDraw(0.,Cy,Wz,Wx,Cy,Wz,PlotPV);
	XYZDraw(Wx,Cy,Wz,Wx,Cy,0.,PlotPV);
	XYZDraw(Wx,Cy,0.,0.,Cy,0.,PlotPV);
}
#endif
#if	XYZ==3
		if (C>0 && C<Nz)
{
	double	Cz=uZ*(double)C;

	XYZDraw(0.,0.,Cz,Wx,0.,Cz,PlotPV);
	XYZDraw(Wx,0.,Cz,Wx,Wy,Cz,PlotPV);
	XYZDraw(Wx,Wy,Cz,0.,Wy,Cz,PlotPV);
	XYZDraw(0.,Wy,Cz,0.,0.,Cz,PlotPV);
}
#endif
	    }
	    XYZTerm();

	    if (gray==NULL)
		StoreGIF_RGB(file,0,0,size,size,r,g,b);
	    else
		StoreGIF(file,0,0,size,size,NULL,NULL,NULL,r);
	}
}
	(void)TermStoreGIF(file); if (file!=stdout) (void)fclose(file);

	return 0;
}
