
#include <stdio.h>
#include <stdlib.h>

extern void	Error();

#define BYTE	unsigned char

static size_t	fwrite_reverse(data,size,memb,file)
BYTE		*data;
size_t		size,memb;
FILE		*file;
{
	int	m,s;

	for (m=0; m<memb; m++) {
	    data+=size;
	    for (s=1; s<=size; s++)
		if (putc((int)data[-s],file)==EOF) goto Break;
	}
Break:;
	return (size_t)m;
}

#define WORD	unsigned short

static double	uX=1.,uY=1.,uZ=1.;
static FILE	*STL;
static size_t	(*fWrite)();

static void	Print(nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk,R,G,B)
int		nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk,R,G,B;
{
	float	xyz[12];
	WORD	flag=(WORD)(0x8000|((R&248)<<7)|((G&248)<<2)|((B&248)>>3));

	xyz[0]=   (float)nx; xyz[ 1]=   (float)ny; xyz[ 2]=   (float)nz;
	xyz[3]=uX*(float)xi; xyz[ 4]=uY*(float)yi; xyz[ 5]=uZ*(float)zi;
	xyz[6]=uX*(float)xj; xyz[ 7]=uY*(float)yj; xyz[ 8]=uZ*(float)zj;
	xyz[9]=uX*(float)xk; xyz[10]=uY*(float)yk; xyz[11]=uZ*(float)zk;
	if ((*fWrite)(xyz,sizeof(float),(size_t)12,STL)!=12 ||
	    (*fWrite)(&flag,sizeof(WORD),(size_t)1,STL)!=1)
	    Error("STL data write error.");
}

typedef struct {
		WORD	hi,vi,hj,vj,hk,vk,d;
		BYTE	R,G,B;
#ifdef	__BORLANDC__
/*
		BYTE	pad[3];
*/
#endif
	} Trigon;

#define PRINT(function,nx,ny,nz,xi,yi,zi,xj,yj,zj,xk,yk,zk)	\
	static void	function(t) \
	Trigon		*t; \
	{ \
		Print(nx,ny,nz,(int)t->xi,(int)t->yi,(int)t->zi, \
			       (int)t->xj,(int)t->yj,(int)t->zj, \
			       (int)t->xk,(int)t->yk,(int)t->zk, \
			       (int)t->R,(int)t->G,(int)t->B); \
	}

PRINT(Mx,-1, 0, 0,d ,hi,vi,d ,hj,vj,d ,hk,vk)
PRINT(Px, 1, 0, 0,d ,hi,vi,d ,hj,vj,d ,hk,vk)
PRINT(My, 0,-1, 0,vi,d ,hi,vj,d ,hj,vk,d ,hk)
PRINT(Py, 0, 1, 0,vi,d ,hi,vj,d ,hj,vk,d ,hk)
PRINT(Mz, 0, 0,-1,hi,vi,d ,hj,vj,d ,hk,vk,d )
PRINT(Pz, 0, 0, 1,hi,vi,d ,hj,vj,d ,hk,vk,d )

#define COUNT	unsigned int

static FILE	*Open(prefix,suffix,path)
char		*prefix,*suffix,*path;
{
	FILE	*file;

	(void)sprintf(path,"%s.%s_m",prefix,suffix);
	if ((file=fopen(path,"rb"))==NULL) {
	    (void)fputs(path,stderr); Error(" : VM file not found.");
	}
	return file;
}

static COUNT	Count(file,path)
FILE		*file;
char		*path;
{
	long	offset;

	if (fseek(file,0L,SEEK_END) || (offset=ftell(file))%sizeof(Trigon)!=0 ||
	    fseek(file,0L,SEEK_SET)) {
	    (void)fputs(path,stderr); Error(" : bad number of trigons.");
	}
	return (COUNT)(offset/sizeof(Trigon));
}

static void	Scan(file,path,count,print)
FILE		*file;
char		*path;
COUNT		count;
void		(*print)();
{
	Trigon	t;

	for (; count; count--) {
	    if (fread(&t,sizeof(Trigon),(size_t)1,file)!=1) {
		(void)fputs(path,stderr); Error(" : trigon data read error.");
	    }
	    (*print)(&t);
	}
	(void)fclose(file);
}

#define PATH_LEN	1024

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*Fmx,*Fpx,*Fmy,*Fpy,*Fmz,*Fpz;
	char	mx[PATH_LEN],px[PATH_LEN],
		my[PATH_LEN],py[PATH_LEN],
		mz[PATH_LEN],pz[PATH_LEN];
	COUNT	Cmx,Cpx,Cmy,Cpy,Cmz,Cpz,facets;

	if (argc!=3 && argc!=6)
	    Error("usage : vm_m_stl VM {unitX unitY unitZ} STL");

	facets=(Cmx=Count(Fmx=Open(argv[1],"mx",mx),mx))+
	       (Cpx=Count(Fpx=Open(argv[1],"px",px),px))+
	       (Cmy=Count(Fmy=Open(argv[1],"my",my),my))+
	       (Cpy=Count(Fpy=Open(argv[1],"py",py),py))+
	       (Cmz=Count(Fmz=Open(argv[1],"mz",mz),mz))+
	       (Cpz=Count(Fpz=Open(argv[1],"pz",pz),pz));

	if (argc==6) {
	    uX=atof(argv[2]); uY=atof(argv[3]); uZ=atof(argv[4]);
	}
	if (argv[argc-1][0]=='-')
	    STL=stdout;
	else
	    if ((STL=fopen(argv[2],"wb"))==NULL) Error("STL file not open.");
{
	int	i;
	short	s=0x1234;
	char	*c=(char *)&s;

	for (i=0; i<80; i++)
	    if (putc(0,STL)==EOF) Error("STL header write error.");

	fWrite=(*c==(s&0xff))?fwrite:fwrite_reverse;
}
	if ((*fWrite)(&facets,sizeof(facets),(size_t)1,STL)!=1)
	    Error("STL facets write error.");

	Scan(Fmx,mx,Cmx,Mx);
	Scan(Fpx,px,Cpx,Px);
	Scan(Fmy,my,Cmy,My);
	Scan(Fpy,py,Cpy,Py);
	Scan(Fmz,mz,Cmz,Mz);
	Scan(Fpz,pz,Cpz,Pz); if (STL!=stdout) (void)fclose(STL); return 0;
}
